/*
 * @(#)$Id$
 *
 * Copyright (c)2000 Regents of the University of California
 *
 */


package precache.util.graph;

import java.util.Set;
import java.util.AbstractSet;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Collections;
import java.util.Iterator;


/**
 *  The <code>GraphRep</code> description.
 *  @version $Revision$, $Date$
 *  @author Kristian Kvilekval
 */
interface GraphRep {

    /* Private/Public Variables */

    public void addNode (Object o);
    public void remNode (Object o);
    public boolean isNode (Object v1);

//     public void addEdge (Object from, Object to);
//     public void remEdge (Object from, Object to);
//     public boolean hasEdge (Object from, Object to);

//     public int countNodes ();
    public Set getNodes();
    public Set getAdjacentNodes (Object o);

    public void addEdge (Edge e);
    public void remEdge (Edge e);
    public boolean hasEdge (Edge e);

//     public int countEdges ();
    public Set getEdges();
    public Set getAdjacentEdges (Object o);
};

class AdjacencyListGraphRep implements GraphRep {

    protected HashMap adjlist;

    protected  HashSet getEdgeSet (Object o) {
        HashSet adj = (HashSet)adjlist.get (o);
        if (adj == null) {
             adjlist.put (o, adj = new HashSet ());
        }
        return adj;
    }
    
    AdjacencyListGraphRep () {
        adjlist = new HashMap ();
    }
    public void addNode (Object o) {
        if (!adjlist.containsKey(o)) 
            getEdgeSet(o);
    }
    public void remNode (Object o) {
        adjlist.remove(o);
    }
    public void addEdge (Object from, Object to) {
        getEdgeSet(from).add (to);
    }
    public void remEdge (Object from, Object to) {
        getEdgeSet(from).remove (to);
    }
    public boolean isNode (Object v1) {
        return adjlist.containsKey (v1);
    }
    public boolean hasEdge (Object from, Object to) {
        return getEdgeSet(from).contains(to);
    }
    public int countNodes () {
        return adjlist.size();
    }
    public Set getNodes() {
        return adjlist.keySet();
    }
    public Set getAdjacentNodes (Object o){
        return (Set)getEdgeSet (o);
    }
    public void addEdge (Edge e) {
        getEdgeSet(e.getVertexA()).add (e.getVertexB());
        
    }
    public void remEdge (Edge e) {
        getEdgeSet(e.getVertexA()).remove (e.getVertexB());
    }
    public boolean hasEdge (Edge e) {
        return  getEdgeSet(e.getVertexA()).contains (e.getVertexB());
    }

    private  class EdgeIterator implements Iterator {
        private Iterator nodeIterator;
        private Object   nodeA;
        private Iterator otherNode;
        private Object   nodeB;
        private Object   edge;
        

        private void nextEdge() {
            if (otherNode != null && otherNode.hasNext()) {
                nodeB = otherNode.next();
                edge = new EdgeImpl (nodeA, nodeB);
            }
        }
        private void primeNext () {
            nextEdge ();
            while (nodeIterator.hasNext() && edge == null) {
                nodeA = nodeIterator.next();
                otherNode = getEdgeSet (nodeA).iterator();
                nextEdge();
            }
            
        }
        EdgeIterator (Iterator nodes) {
            nodeIterator = nodes;
            primeNext();
        }

        public boolean hasNext() {
            return edge != null;
        }
        public Object next() {
            if (edge != null) {
                Object e =  edge;
                edge = null;
                primeNext();
                return e;
            }
            throw new java.util.NoSuchElementException();
        }
        public void remove (){
            throw new UnsupportedOperationException();
        }
    }

    public Set getEdges() {
        return new AbstractSet () {
                public Iterator iterator() {
                    return new EdgeIterator (adjlist.keySet().iterator());
                }
                public int size() {
                    Iterator ni = adjlist.keySet().iterator();
                    int count = 0;
                    while (ni.hasNext()){
                        count += getEdgeSet(ni.next()).size();
                    }
                    return count;
                }
            };
    }
    public Set getAdjacentEdges (final Object o) {
        return new AbstractSet () {
                public Iterator iterator() {
                    return new EdgeIterator (Collections.singleton(o).iterator());
                }
                public int size() {
                    return getEdgeSet (o).size();
                }
            };
        
    }

}


/** Store the adjacency matrix for an undirected graph.
 *  We map node to numbers then use the node number to
 *  to index the matix
 *         To
 *      1 2 3 4             Lower triangular (picture is wrong)
 * F  1   x   x
 * r  2     x
 * o  3       x
 * m  4       x
 */
class AdjacencyMatrixGraphRep implements GraphRep {

    protected HashMap  nodeIndex;
    protected Object[] nodes;
    protected int[]    adjmatrix;
    protected int      maxnodes;

    protected int getNodeIndex (Object o) {
        return ((Integer)nodeIndex.get (o)).intValue();
    }
    /** make  a <-> b .. B must be less than A */
    int getIndex (int noda, int nodeb) {
        if (nodeb < noda) { int tmp = nodeb; nodeb=noda; noda=tmp; }
        return ((noda * (noda+1)))/2 + nodeb;
    }

    int getMatrixIndex (Object from, Object to) {
        int vfrom = getNodeIndex (from);
        int vto = getNodeIndex (to);
        return getIndex(vfrom, vto);
    }

    AdjacencyMatrixGraphRep(){
    }
    AdjacencyMatrixGraphRep(int maxnodes) {
        this.maxnodes = maxnodes;
        this.nodeIndex = new HashMap();
        this.nodes = new Object[maxnodes];
        this.adjmatrix = new int[maxnodes*(maxnodes+1)/2];
    }    
    public   void addNode (Object o) {
        if (!nodeIndex.containsKey(o)){
            int index = nodeIndex.size();
            nodeIndex.put (o, new Integer (index));
            nodes[index] = o;
        }
    }
    public void remNode (Object o) {
        // throw unimplemented.
    }
    public void addEdge (Object from, Object to) {
        int index  = getMatrixIndex (from, to);
        adjmatrix[index] = 1;
    }
    public void remEdge (Object from, Object to) {
        int index  = getMatrixIndex (from, to);
        adjmatrix[index] = 1;
    }
    public boolean isNode (Object v1) {
        return false;
    }
    public boolean hasEdge (Object from, Object to) {
        return (adjmatrix[getMatrixIndex(from,to)] != 0);
    }
    public void addEdge (Edge e) {
    }
    public void remEdge (Edge e) {
    }
    public boolean hasEdge (Edge e) {
        return false;
    }

    public int countNodes() {
        return nodes.length;
    }

    public Set getNodes() {
        return nodeIndex.keySet();
    }
    public Set getAdjacentNodes (Object o){
        HashSet adj = new HashSet ();
        // Scan the matrix adding the nodes to the set;
        return adj;
    }


    private class EdgeIterator implements Iterator {
        private int nodeA;
        private int nodeB;

        EdgeIterator () {
            nodeA = 0; nodeB = -1;
            findnext();
        }
        private void findnext () {
            while (nodeA < maxnodes) {
                while (++nodeB < maxnodes) {
                    if (adjmatrix[getIndex(nodeA,nodeB)] == 1) 
                        return;
                }
                nodeA++;
            }
        }
        public boolean hasNext() {
            return (nodeA < maxnodes && nodeB < maxnodes);
        }
        public Object next() {
            if (hasNext())
                return new EdgeImpl (nodes[nodeA], nodes[nodeB]);
            throw new java.util.NoSuchElementException();
        }
        public void remove (){
            throw new java.lang.UnsupportedOperationException("EdgeSet");
        }
    }


    public Set getEdges() {
        return new AbstractSet () {
                public Iterator iterator() {
                    return new EdgeIterator ();
                }
                public int size() {
                    int count = 0;
                    int last = maxnodes*(maxnodes+1)/2;
                    for (int n = 0; n < last; n++) {
                        if (adjmatrix[n] == 1) 
                            count++;
                    }
                    return count;
                }

            };
    }
    public Set getAdjacentEdges (Object o) {
        return null;
    }

}

/** same as  AdjacencyMatrixGraphRep but store the
 *  the full matrix 
 */
class AdjacencyMatrixDirectedGraphRep 
    extends AdjacencyMatrixGraphRep {


    protected int getMatrixIndex (Object from, Object to) {
        int vfrom = getNodeIndex (from);
        int vto = getNodeIndex (to);
        return ((vfrom * maxnodes)) + vto;
    }

    AdjacencyMatrixDirectedGraphRep (int maxcount) {
        super();
        this.maxnodes = maxnodes;
        this.nodeIndex = new HashMap();
        this.nodes = new Object[maxnodes];
        this.adjmatrix = new int[maxnodes*maxnodes];
    }

//     public int countNodes() {
//         return 0;
//     }

//     public Set getNodes() {
//         return null;
//     }
//     public Set getAdjacentNodes (Object o){
//         return null;
//     }

}

/** Support user defined edge types 
 *  Maintain a hash table table of nodes
 *  to edges.  
 */
class AdjacencyListEdgeRep extends  AdjacencyListGraphRep{

    HashMap adjlist;


    public void addNode (Object o) {
        throw new UnsupportedOperationException ("addNode");
    }
    public void remNode (Object o) {
    }
    public void addEdge (Object from, Object to){
    }
    public void remEdge (Object from, Object to) {
    }
    //public boolean isNode (Object v1);
    //public boolean hasEdge (Object from, Object to);

    public void addEdge (Edge e) {
        getEdgeSet (e.getVertexA()).add (e);
        getEdgeSet (e.getVertexB()).add (e);
    }
    public void remEdge (Edge e) {
        getEdgeSet (e.getVertexA()).remove (e);
        getEdgeSet (e.getVertexB()).remove (e);
    }
    public boolean hasEdge (Edge e){
        return getEdgeSet (e.getVertexA()).contains (e);
    }

    //public int countNodes ();
    //public Set getNodes();
    //public Set getAdjacentNodes (Object o);

    public int countEdges () {
        return 0;
    }
    public Set getEdges() {
        return null;
    }
    public Set getAdjacentEdges (Object o) {
        return null;
    }
    
}

/** A representation in which the nodes themselves
 *  maintain links to other nodes.  The Object argument
 *  must support the GrahNode interface 
 */
class GraphNodeRep implements GraphRep {

    HashSet nodes;

    GraphNodeRep() {
        nodes = new HashSet();
    }
    public void addNode (Object o) {
        nodes.add (o);
    }
    public void remNode (Object o) {
        nodes.remove (o);
    }
    public void addEdge (Object from, Object to) {
        ((GraphNode)from).addEdge (to);
    }
    public void remEdge (Object from, Object to){
        ((GraphNode)from).remEdge (to);
    }
    public boolean isNode (Object v1) {
        return nodes.contains (v1);
    }
    public boolean hasEdge (Object from, Object to){
        return ((GraphNode)from).hasEdge (to);
    }

    public void addEdge (Edge e) {
    }
    public void remEdge (Edge e) {
    }
    public boolean hasEdge (Edge e) {
        return false;
    }

    public int countNodes () {
        return nodes.size();
    }
    // Change Name to nodes for consistency
    public Set getNodes() {
        return nodes;
    }
    public Set getAdjacentNodes (Object o){
        return ((GraphNode)o).adjacentNodes();
    }

    public int countEdges () {
        return 0;
    }
    public Set getEdges() {
        return null;
    }
    public Set getAdjacentEdges (Object o) {
        return null;
    }

}


